import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.Frame;
import win32frame.WindowMenu;
import win32frame.StatusBar;
import win32frame.MainWindow ;

import std.stdio;
import std.string;
import std.utf;

import Sound ;

ushort LOWORD(uint l){ return l & 0xffff; }
ushort HIWORD(uint l){ return l >>> 16; }

const SRATE    = 11025 ;
const PLAYTIME =    60 ;

class MainFrame : MainWindow {
  WindowMenu   menuroot ;
  StatusBar    statusbar ;
  
  HFONT        hFont;
  HWAVEOUT     hWave;
  WAVEFORMATEX wfe;
  WAVEHDR      whdr;
  char[]       bWave;
  

  this(string name, string ClassName) {
    winc.style          = 0; //CS_HREDRAW | CS_VREDRAW;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.hIcon          = LoadIcon(null , cast(wchar*)(IDI_APPLICATION));
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.hbrBackground  = CreateSolidBrush(GetSysColor(COLOR_BTNFACE)) ;
    winc.lpszMenuName   = null;
    winc.lpszClassName  = cast(wchar*)(ClassName) ;
    super(name,ClassName) ;
    
  }

  LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      waveOutClose(hWave);
      DeleteObject(hFont);
      PostQuitMessage(0);
      break ;
      
    case WM_CREATE:
      menuroot = new WindowMenu() ;
      menuroot.create() ;
      menuroot.addMenu(         MFT_STRING, "File"      , 0x000 ) ;
      menuroot.addItem( "File", MFT_STRING, "MML Open"  , 0x101 ) ;
      menuroot.addItem( "File", MFT_STRING, "Exit"      , 0x104 ) ;
      SetMenu(hwnd , menuroot.getHMENU());
      
      statusbar = new StatusBar() ;
      statusbar.addStyle(CCS_BOTTOM | SBARS_SIZEGRIP) ;
      addStatusBar(statusbar, "") ;
      
      hFont = CreateFont(
			 14 , 0 , 0 , 0 , FW_REGULAR , FALSE , FALSE , FALSE ,
			 SHIFTJIS_CHARSET , OUT_DEFAULT_PRECIS ,
			 CLIP_DEFAULT_PRECIS , DEFAULT_QUALITY , 
			 VARIABLE_PITCH, null
			 );

      bWave.length = SRATE*PLAYTIME;
      wfe.wFormatTag      = WAVE_FORMAT_PCM ;
      wfe.nChannels       = 1 ;
      wfe.nSamplesPerSec  = SRATE ;
      wfe.nAvgBytesPerSec = SRATE ;
      wfe.wBitsPerSample  = 8 ;
      wfe.nBlockAlign     = wfe.nChannels * wfe.wBitsPerSample /8 ;
      //      waveOutOpen(&hWave , WAVE_MAPPER , &wfe , cast(uint)(hwnd), 0 , CALLBACK_WINDOW);
      
      whdr.lpData         = bWave.ptr;
      whdr.dwBufferLength = SRATE*PLAYTIME;
      whdr.dwFlags        = WHDR_BEGINLOOP | WHDR_ENDLOOP;
      whdr.dwLoops        = 1;
      //      waveOutPrepareHeader(hWave , &whdr , WAVEHDR.sizeof);
      break ;
      
    case MM_WOM_OPEN:
      //    case MM_WOM_DONE:
      waveOutWrite(cast(HWAVEOUT)(wp) , &whdr , WAVEHDR.sizeof);
      break ;
    case MM_WOM_CLOSE:
      //      waveOutUnprepareHeader(cast(HWAVEOUT)(wp) , &whdr , WAVEHDR.sizeof);
      break ;
		
    case WM_MOVE:
      break ;
      
    case WM_SIZE:
      break;
      
    case WM_LBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_LBUTTONDBLCLK:
    case WM_RBUTTONDOWN:
    case WM_RBUTTONUP:
    case WM_RBUTTONDBLCLK:
    case WM_MOUSEMOVE:
      break ;

    case WM_COMMAND:
      switch( LOWORD(wp) ){
      case 0x101 : //MML Open
	string fname ;
	fname = getLoadFileName("MML {*.mml}\0*.mml\0ALL {*.*}\0*.*\0") ;
	if( fname.length != 0 ){
	  waveOutReset(hWave) ;
	  waveOutClose(hWave);
	  Sound snd = new Sound() ;
	  snd.readMML(fname) ;
	  snd.rendWave(bWave,SRATE) ;
	  waveOutOpen(&hWave , WAVE_MAPPER , &wfe , cast(uint)(hwnd), 0 , CALLBACK_WINDOW);
	  waveOutPrepareHeader(hWave , &whdr , WAVEHDR.sizeof);
	}
	break ;
      case 0x104 : //Exit
	SendMessage( hwnd, WM_CLOSE , 0,0 ) ;
	break ;
      default :	break ;
      }
      break;
      
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  string getLoadFileName(string filematch){
    OPENFILENAME ofn ;
    char[]       fname ;
    wchar[]      wfname ;
    string       retstr ;

    fname.length=MAX_PATH ;
    
    ofn.lStructSize = OPENFILENAME.sizeof ;
    ofn.hwndOwner = getHWND() ;
    ofn.lpstrFilter = cast(wchar*)(toUTF16(filematch)) ;
    ofn.lpstrCustomFilter = null ;
    //    ofn.nMaxCustFilter = 256 ;
    //    ofn.nFilterIndex = 0 ;
    ofn.lpstrFile = cast(wchar*)(fname.ptr) ;
    ofn.nMaxFile = MAX_PATH ;
    ofn.Flags = OFN_FILEMUSTEXIST ;

    GetOpenFileName(&ofn) ;

    // 何故かwchar[]型をofn.lpstrFileに指定するとファイル選択ダイアログが開かない。
    // (1)char[]型のfnameに得られた UTF16文字 をwchar[]型の wfnameに移す。
    // (2)無効文字が0xffff(0x00ffではない)で埋まるので、長さを詰める
    // (3)std.utf.toUTF8でUTF8文字に変換したretstrを内部では使用する。
    // という操作が必要らしい。
    
    wfname.length=MAX_PATH ;
    for( int i=0 ; i<fname.length ; i+=2 ){ //char[]→wchar[] 移動
      wfname[i/2] = (cast(ushort)(fname[i+1])<<8 | cast(ushort)(fname[i])) ;
    }
    for( int i=0 ; i<wfname.length ; i++ ){// 0xffff文字の手前で切る
      if( wfname[i]==0xffff ){
	wfname.length=i ;
	break ;
      }
    }
    retstr = std.utf.toUTF8(wfname.idup) ;
    return(retstr) ;
  }
}
